/*
 Copyright (c) 2013 aizo ag, Zurich, Switzerland

 Author: Christain Hitz <christian.hitz@aizo.com>

 This file is part of dS Assistant.

 dS-Assistant is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 dS-Assistant is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

 NOTE: all APIs, especially the JSON API, offered by this application
 are subject to change without further notice.
 */

#include "device-config.h"
#include <algorithm>
#include "util.h"
#include <cerrno>
#include <json.h>
#include <fstream>
#include <string>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>
#include "log.h"
#include "config_manager.h"

namespace dSA {

SceneConfig::SceneConfig(int dimmTime, int ledConfig, bool flash, bool special,
                         bool localprio, bool dontcare)
        : m_dimmTime(dimmTime),
          m_ledConfig(ledConfig),
          m_flash(flash),
          m_special(special),
          m_localprio(localprio),
          m_dontcare(dontcare)
{
}

bool SceneConfig::isEqual(const boost::shared_ptr<SceneConfig>& config) const
{
    if ((m_dimmTime == config->getDimmTime())
        && (m_ledConfig == config->getLedConfig())
        && (m_flash == config->isFlash())
        && (m_special == config->isSpecial())
        && (m_localprio == config->isLocalprio())
        && (m_dontcare == config->isDontcare()))
    {
        return true;
    }
    return false;
}

boost::shared_ptr<ParameterSpec> SceneConfig::getParameterSpec() const
{
    boost::shared_ptr<ParameterSpec> reg(new ParameterSpec());
    reg->bank = 128;
    reg->offset = 0;
    reg->value = ((m_dimmTime & 0x3) << 6)
                 | ((m_ledConfig & 0x3) << 4)
                 | (m_flash << 3)
                 | (m_special << 2)
                 | (m_localprio << 1)
                 | (m_dontcare);
    reg->mask = 0xff;
    return reg;
}

Scene::Scene(int sceneNumber, bool active, boost::shared_ptr<SceneConfig>& sceneConfig)
        : m_sceneNumber(sceneNumber),
          m_active(active),
          m_sceneConfig(sceneConfig)
{
}

bool Scene::isConfigEqual(const Scene* scene) const
{
    if ( !(m_active ^ scene->isActive()))
    {
        return true;
    }
    if (m_sceneNumber == scene->getSceneNumber())
    {
        return m_sceneConfig->isEqual(scene->getSceneConfig());
    }
    return false;
}

std::vector<boost::shared_ptr<ParameterSpec> > Scene::parameterChangesConfig(
        const Scene* scene) const
{
    std::vector<boost::shared_ptr<ParameterSpec> > res;
    if ((scene == NULL) || !Scene::isConfigEqual(scene))
    {
        std::vector<boost::shared_ptr<ParameterSpec> > params =
                Scene::getParameterSpec();
        res.insert(res.end(), params.begin(), params.end());
    }
    return res;
}

std::vector<boost::shared_ptr<ParameterSpec> > Scene::getParameterSpec() const
{
    std::vector<boost::shared_ptr<ParameterSpec> > regs;

    boost::shared_ptr<ParameterSpec> config =
            getSceneConfig()->getParameterSpec();
    config->offset = 128 + getSceneNumber();
    regs.push_back(config);

    return regs;
}

LightScene::LightScene(int sceneNumber, bool active, uint8_t value,
                       boost::shared_ptr<SceneConfig>& sceneConfig)
        : Scene(sceneNumber, active, sceneConfig),
          m_value(value)
{
}

bool LightScene::isEqual(const Scene* scene) const
{
    if (const LightScene* other = dynamic_cast<const LightScene*>(scene))
    {
        if ( !(m_active ^ other->isActive()))
        {
            return true;
        }
        if (m_value == other->getValue())
        {
            return Scene::isConfigEqual(other);
        }
    }
    return false;
}

std::vector<boost::shared_ptr<ParameterSpec> > LightScene::parameterChanges(
        const Scene* scene) const
{
    std::vector<boost::shared_ptr<ParameterSpec> > res;
    if (scene == NULL)
    {
        std::vector<boost::shared_ptr<ParameterSpec> > params =
                LightScene::getParameterSpec();
        res.insert(res.end(), params.begin(), params.end());

        params = Scene::parameterChangesConfig(scene);
        res.insert(res.end(), params.begin(), params.end());
    }
    else if (const LightScene* other = dynamic_cast<const LightScene*>(scene))
    {
        if ((m_active == true && !other->isActive()))
        {
            if (m_value != other->getValue())
            {
                std::vector<boost::shared_ptr<ParameterSpec> > params =
                        LightScene::getParameterSpec();
                res.insert(res.end(), params.begin(), params.end());
            }
            std::vector<boost::shared_ptr<ParameterSpec> > params =
                    Scene::parameterChangesConfig(scene);
            res.insert(res.end(), params.begin(), params.end());
        }
    }
    return res;
}

std::vector<boost::shared_ptr<ParameterSpec> > LightScene::getParameterSpec() const
{
    std::vector<boost::shared_ptr<ParameterSpec> > regs;

    boost::shared_ptr<ParameterSpec> value(new ParameterSpec());
    value->bank = 128;
    value->offset = getSceneNumber();
    value->value = getValue();
    value->mask = 0xFF;
    regs.push_back(value);

    return regs;
}

MotorScene::MotorScene(int sceneNumber, bool active, uint16_t value,
                       boost::shared_ptr<SceneConfig>& sceneConfig)
        : Scene(sceneNumber, active, sceneConfig),
          m_value(value)
{
}

bool MotorScene::isEqual(const Scene* scene) const
{
    if (const MotorScene* other = dynamic_cast<const MotorScene*>(scene))
    {
        if ( !(m_active ^ other->isActive()))
        {
            return true;
        }
        if (m_value == other->getValue())
        {
            return Scene::isConfigEqual(other);
        }
    }
    return false;
}

std::vector<boost::shared_ptr<ParameterSpec> > MotorScene::parameterChanges(
        const Scene* scene) const
{
    std::vector<boost::shared_ptr<ParameterSpec> > res;
    if (scene == NULL)
    {
        std::vector<boost::shared_ptr<ParameterSpec> > params =
                MotorScene::getParameterSpec();
        res.insert(res.end(), params.begin(), params.end());

        params = Scene::parameterChangesConfig(scene);
        res.insert(res.end(), params.begin(), params.end());
    }
    else if (const MotorScene* other = dynamic_cast<const MotorScene*>(scene))
    {
        if ((m_active == true && !other->isActive()))
        {
            if (m_value != other->getValue())
            {
                std::vector<boost::shared_ptr<ParameterSpec> > params =
                        MotorScene::getParameterSpec();
                res.insert(res.end(), params.begin(), params.end());
            }
            std::vector<boost::shared_ptr<ParameterSpec> > params =
                    Scene::parameterChangesConfig(scene);
            res.insert(res.end(), params.begin(), params.end());
        }
    }
    return res;
}

std::vector<boost::shared_ptr<ParameterSpec> > MotorScene::getParameterSpec() const
{
    std::vector<boost::shared_ptr<ParameterSpec> > regs;

    boost::shared_ptr<ParameterSpec> value(new ParameterSpec());
    value->bank = 128;
    value->offset = getSceneNumber();
    value->value = (getValue() & 0x0FF0) >> 4;
    value->mask = 0xFF;
    regs.push_back(value);

    boost::shared_ptr<ParameterSpec> value_lo(new ParameterSpec());
    value_lo->bank = 129;
    value_lo->offset = getSceneNumber() / 2;
    value_lo->value = (getValue() & 0x000F) << ((getSceneNumber() % 2) * 4);
    value_lo->mask = 0x0F << ((getSceneNumber() % 2) * 4);
    regs.push_back(value_lo);

    return regs;
}

JalusieScene::JalusieScene(int sceneNumber, bool active, uint16_t value, uint8_t angle,
                           boost::shared_ptr<SceneConfig>& sceneConfig)
        : MotorScene(sceneNumber, active, value, sceneConfig),
          m_angle(angle)
{
}

bool JalusieScene::isEqual(const Scene* scene) const
{
    if (const JalusieScene* other = dynamic_cast<const JalusieScene*>(scene))
    {
        if (m_angle == other->getAngle())
        {
            return MotorScene::isEqual(other);
        }
    }
    return false;
}

std::vector<boost::shared_ptr<ParameterSpec> > JalusieScene::parameterChanges(
        const Scene* scene) const
{
    std::vector<boost::shared_ptr<ParameterSpec> > res;
    if (scene == NULL)
    {
        std::vector<boost::shared_ptr<ParameterSpec> > params =
                getParameterSpec();
        res.insert(res.end(), params.begin(), params.end());
        params = MotorScene::parameterChanges(scene);
        res.insert(res.end(), params.begin(), params.end());
    }
    else if (const JalusieScene* other =
            dynamic_cast<const JalusieScene*>(scene))
    {
        if ((m_active == true && !other->isActive()))
        {
            if (m_angle != other->getAngle())
            {
                std::vector<boost::shared_ptr<ParameterSpec> > params =
                        getParameterSpec();
                res.insert(res.end(), params.begin(), params.end());
            }
            std::vector<boost::shared_ptr<ParameterSpec> > params =
                    MotorScene::parameterChanges(scene);
            res.insert(res.end(), params.begin(), params.end());
        }
    }
    return res;
}

std::vector<boost::shared_ptr<ParameterSpec> > JalusieScene::getParameterSpec() const
{
    std::vector<boost::shared_ptr<ParameterSpec> > regs;

    boost::shared_ptr<ParameterSpec> angle(new ParameterSpec());
    angle->bank = 130;
    angle->offset = getSceneNumber();
    angle->value = getAngle();
    angle->mask = 0xFF;
    regs.push_back(angle);

    return regs;
}

DeviceConfig::DeviceConfig(uint8_t colorId, uint16_t productId,
                           uint16_t fwVersion, const std::string& name,
                           int version)
        : m_version(version),
          m_colorId(colorId),
          m_productId(productId),
          m_fwVersion(fwVersion),
          m_name(name)
{
}

void DeviceConfig::addScene(boost::shared_ptr<Scene> scene)
{
    m_vScene.push_back(scene);
}

void DeviceConfig::addParameter(boost::shared_ptr<ParameterSpec> parameter)
{
    m_vParameters.push_back(parameter);
}

bool ParameterSpecPred(const boost::shared_ptr<ParameterSpec>& a,
                       const boost::shared_ptr<ParameterSpec>& b)
{
    return (a->bank != b->bank) ? (a->bank < b->bank) : (a->offset < b->offset);
}

std::vector<boost::shared_ptr<ParameterSpec> > DeviceConfig::parameterChanges(
        const DeviceConfig& other)
{
    const std::vector<boost::shared_ptr<Scene> >* old = &m_vScene;
    const std::vector<boost::shared_ptr<Scene> >* current = &other.getScenes();
    const std::vector<boost::shared_ptr<ParameterSpec> >* oldParams = &m_vParameters;
    const std::vector<boost::shared_ptr<ParameterSpec> >* currentParams = &other.getParameters();
    std::vector<boost::shared_ptr<ParameterSpec> > vChangedParams;

    if ((other.getVersion() == m_version)
        || (other.getColorId() != m_colorId)
        || (other.getProductId() != m_productId))
    {
        return vChangedParams;
    }
    else if (other.getVersion() < m_version)
    {
        std::swap(old, current);
        std::swap(oldParams, currentParams);
    }

    std::vector<boost::shared_ptr<Scene> >::const_iterator oldIter =
            old->begin();
    std::vector<boost::shared_ptr<Scene> >::const_iterator curIter =
            current->begin();
    while (curIter < current->end())
    {
        if ((oldIter >= old->end())
            || ((*curIter)->getSceneNumber() != (*oldIter)->getSceneNumber()))
        {
            /* new scene, not available in old config */
            std::vector<boost::shared_ptr<ParameterSpec> > res =
                    (*curIter)->parameterChanges(NULL);
            vChangedParams.insert(vChangedParams.end(), res.begin(), res.end());
        }
        else
        {
            std::vector<boost::shared_ptr<ParameterSpec> > res =
                    (*curIter)->parameterChanges(oldIter->get());
            vChangedParams.insert(vChangedParams.end(), res.begin(), res.end());
            ++oldIter;
        }
        ++curIter;
    }

    std::vector<boost::shared_ptr<ParameterSpec> >::const_iterator oldParamsIter =
            oldParams->begin();
    std::vector<boost::shared_ptr<ParameterSpec> >::const_iterator curParamsIter =
            currentParams->begin();
    while (curParamsIter < currentParams->end())
    {
        if ((oldParamsIter >= oldParams->end())
            || ((*curParamsIter)->bank != (*oldParamsIter)->bank)
            || ((*curParamsIter)->offset != (*oldParamsIter)->offset))
        {
            /* new parameter, not available in old config */
            vChangedParams.push_back(*curParamsIter);
        }
        else if (((*curParamsIter)->bank == (*oldParamsIter)->bank) &&
                 ((*curParamsIter)->offset == (*oldParamsIter)->offset) &&
                 ((*curParamsIter)->value != (*oldParamsIter)->value))
        {
            vChangedParams.push_back(*curParamsIter);
            ++oldParamsIter;
        }
        ++curParamsIter;
    }

    std::sort(vChangedParams.begin(), vChangedParams.end(), ParameterSpecPred);
    return vChangedParams;
}

std::vector<int> DeviceConfig::sceneChanges(const DeviceConfig& other)
{
    const std::vector<boost::shared_ptr<Scene> >* old = &m_vScene;
    const std::vector<boost::shared_ptr<Scene> >* current = &other.getScenes();
    const std::vector<boost::shared_ptr<ParameterSpec> >* oldParams = &m_vParameters;
    const std::vector<boost::shared_ptr<ParameterSpec> >* currentParams = &other.getParameters();
    std::vector<int> vChangedScenes;

    if ((other.getVersion() == m_version)
        || (other.getColorId() != m_colorId)
        || (other.getProductId() != m_productId))
    {
        return vChangedScenes;
    }
    else if (other.getVersion() < m_version)
    {
        std::swap(old, current);
        std::swap(oldParams, currentParams);
    }

    std::vector<boost::shared_ptr<Scene> >::const_iterator oldIter =
            old->begin();
    std::vector<boost::shared_ptr<Scene> >::const_iterator curIter =
            current->begin();
    while (curIter < current->end())
    {
        if ((oldIter >= old->end())
            || ((*curIter)->getSceneNumber() != (*oldIter)->getSceneNumber()))
        {
            /* new scene, not available in old config */
            vChangedScenes.push_back((*curIter)->getSceneNumber());
        }
        else
        {
            if (!(*curIter)->isEqual(oldIter->get()))
            {
                vChangedScenes.push_back((*curIter)->getSceneNumber());
            }
            ++oldIter;
        }
        ++curIter;
    }

    std::vector<boost::shared_ptr<ParameterSpec> >::const_iterator oldParamsIter =
            oldParams->begin();
    std::vector<boost::shared_ptr<ParameterSpec> >::const_iterator curParamsIter =
            currentParams->begin();
    while (curParamsIter < currentParams->end())
    {
        if ((oldParamsIter >= oldParams->end())
            || ((*curParamsIter)->bank != (*oldParamsIter)->bank)
            || ((*curParamsIter)->offset != (*oldParamsIter)->offset))
        {
            /* new parameter, not available in old config */
            vChangedScenes.push_back(-1);
            break;
        }
        else if (((*curParamsIter)->bank == (*oldParamsIter)->bank) &&
                 ((*curParamsIter)->offset == (*oldParamsIter)->offset) &&
                 ((*curParamsIter)->value != (*oldParamsIter)->value))
        {
            vChangedScenes.push_back(-1);
            break;
        }
        ++curParamsIter;
    }
    return vChangedScenes;
}

std::string DeviceConfigReader::get_file_contents(const std::string& filename)
{
    std::ifstream in(filename.c_str(), std::ios::in | std::ios::binary);
    if (in)
    {
        std::string contents;
        in.seekg(0, std::ios::end);
        contents.resize(in.tellg());
        in.seekg(0, std::ios::beg);
        in.read(&contents[0], contents.size());
        in.close();
        return (contents);
    }
    throw(errno);
}

boost::shared_ptr<DeviceConfig> DeviceConfigReader::read(const Device& device,
                                                         int version)
{
    std::string dir = ConfigManager::getInstance()->getOption(CFG_DEVICE_CONFIG_DIR);
    std::string devPrefix = device.getHardwareInfoString();
    std::string filePattern;
    std::string fileName;
    std::vector<std::string> fileList;
    std::vector<std::string> fileListRevision;

    filePattern = dir + "/" + devPrefix + "-*.json";
    fileList = fileutil::glob(filePattern, 0);
    if (fileList.empty())
    {
        return boost::shared_ptr<DeviceConfig> ();
    }

    FILE_LOG(logINFO) << "DeviceConfigReader: " << util::str(device.getDSUID()) <<
            ", Pattern: " << filePattern <<
            ", FW: " << util::hex(device.getVersion(), 4) <<
            ", Version: " << version;

    try
    {
        // find exact match for firmware revision
        for (std::vector<std::string>::iterator it = fileList.begin(); it != fileList.end(); it++)
        {
            boost::regex expr(
                    ".*/" + devPrefix + "-" + util::hex(device.getVersion(), 4).substr(2) + "-" + "[a-f0-9]{2}.json");
            if (boost::regex_match(*it, expr))
            {
                FILE_LOG(logINFO) << "DeviceConfigReader:  match exact revision: " << *it;
                fileListRevision.push_back(*it);
            }
        }

        // find best match for firmware revision
        if (fileListRevision.empty())
        {
            uint16_t rev = device.getVersion();
            for (; rev > 0x300; rev--)
            {
                for (std::vector<std::string>::iterator it = fileList.begin(); it != fileList.end(); it++)
                {
                    boost::regex expr(".*/" + devPrefix + "-" + util::hex(rev, 4).substr(2) + "-" + "[a-f0-9]{2}.json");
                    if (boost::regex_match(*it, expr))
                    {
                        FILE_LOG(logINFO) << "DeviceConfigReader:  match nearest revision: " << *it;
                        fileListRevision.push_back(*it);
                    }
                }
            }
        }

        // try to match an exact config revision
        if ((version > 0) && !fileList.empty())
        {
            for (std::vector<std::string>::iterator it = fileList.begin(); it != fileList.end(); it++)
            {
                boost::regex expr(
                        ".*/" + devPrefix + "-" + "[a-f0-9]{4}" + "-" + util::hex(version, 2).substr(2) + ".json");
                if (boost::regex_match(*it, expr))
                {
                    FILE_LOG(logINFO) << "DeviceConfigReader:  match exact version: " << *it;
                    fileName = *it;
                }
            }
        }

        // version < 0 -> find the newest available config file for this device type
        if ((version < 0) && !fileList.empty())
        {
            // grab the last the one of the glob'ed list which is already sorted
            fileName = fileList[fileList.size() - 1];
            FILE_LOG(logINFO) << "DeviceConfigReader:  match newest: " << fileName;

            boost::regex expr(".*/" + devPrefix + "-" + "[a-f0-9]{4}" + "-" + "([a-f0-9]{2})" + ".json",
                    boost::regex::extended);
            boost::smatch result;
            if (boost::regex_search(fileName, result, expr))
            {
                try
                {
                    version = util::str2int(result[1]);
                    FILE_LOG(logINFO) << "DeviceConfigReader:  extracted version: " << version << " from " << fileName;
                }
                catch (Exception& e)
                {
                    FILE_LOG(logERROR) << "DeviceConfigReader: "
                            "cannot extract config version from "
                            << fileName << ": " << e.getMessage();
                }
            }
        }

    }
    catch (std::exception& e)
    {
        FILE_LOG(logERROR) << "DeviceConfigReader: "
                "error filtering config files: " << e.what();
    }

    // if version matching was not possible take the best match against the fw revision
    if (fileName.empty() && !fileListRevision.empty())
    {
        // the first entry is the best match and closest fw revision
        fileName = fileListRevision[0];
    }

    if (!fileutil::exists(fileName))
    {
        return boost::shared_ptr<DeviceConfig> ();
    }

    FILE_LOG(logINFO) << "DeviceConfigReader:  using config file " << fileName;

    boost::shared_ptr<DeviceConfig> res(
            new DeviceConfig(device.getStandardColor(), device.getProductId(),
                             device.getVersion(),
                             device.getHardwareInfoString(), version));


    std::string rawConfig;
    try
    {
        rawConfig = get_file_contents(fileName);
    } catch (...)
    {
        return res;
    }
    struct json_tokener* tok;

    tok = json_tokener_new();
    json_object* json_request = json_tokener_parse_ex(tok, rawConfig.c_str(),
                                                      -1);

    if (tok->err == json_tokener_success)
    {
        std::string name;
        json_object* obj;
        if (json_object_object_get_ex(json_request, "productname", &obj))
        {
            name = json_object_get_string(obj);
        }
        if (name != device.getHardwareInfoString())
        {
            return res;
        }

        SceneType sceneType = LIGHT_SCENE;
        if (json_object_object_get_ex(json_request, "type", &obj))
        {
            sceneType = (SceneType) json_object_get_int(obj);
        }
        else
        {
            /* malformed JSON */
            return res;
        }

        json_object* scenetable;
        if (json_object_object_get_ex(json_request, "scenetable", &scenetable))
        {
            for (int i = 0; i < json_object_array_length(scenetable); ++i)
            {
                json_object* scene = json_object_array_get_idx(scenetable, i);
                if (scene != NULL)
                {
                    int sceneNr = 0;
                    if (json_object_object_get_ex(scene, "scene", &obj))
                    {
                        sceneNr = json_object_get_int(obj);
                    }

                    int value = 0;
                    if (json_object_object_get_ex(scene, "value", &obj))
                    {
                        value = json_object_get_int(obj);
                    }

                    int ledconfig = 1;
                    if (json_object_object_get_ex(scene, "ledconfig", &obj))
                    {
                        ledconfig = json_object_get_int(obj);
                    }

                    int dimconfig = 1;
                    if (json_object_object_get_ex(scene, "dimconfig", &obj))
                    {
                        dimconfig = json_object_get_int(obj);
                    }

                    bool dontcare = false;
                    if (json_object_object_get_ex(scene, "dontcare", &obj))
                    {
                        dontcare = (json_object_get_int(obj) == 1);
                    }

                    bool localprio = false;
                    if (json_object_object_get_ex(scene, "localprio", &obj))
                    {
                        localprio = (json_object_get_int(obj) == 1);
                    }

                    bool special = false;
                    if (json_object_object_get_ex(scene, "special", &obj))
                    {
                        special = (json_object_get_int(obj) == 1);
                    }

                    bool flash = false;
                    if (json_object_object_get_ex(scene, "flash", &obj))
                    {
                        flash = (json_object_get_int(obj) == 1);
                    }

                    int angle = 0;
                    if (sceneType == JALUSIE_SCENE)
                    {
                        if (json_object_object_get_ex(scene, "angle", &obj))
                        {
                            angle = json_object_get_int(obj);
                        }
                    }

                    bool active = false;
                    if (json_object_object_get_ex(scene, "active", &obj))
                    {
                        active = (json_object_get_int(obj) == 1);
                    }

                    boost::shared_ptr<SceneConfig> config(
                            new SceneConfig(dimconfig, ledconfig, flash, special, localprio, dontcare));
                    boost::shared_ptr<Scene> pScene;
                    switch (sceneType)
                    {
                        case LIGHT_SCENE:
                        {
                            pScene = boost::shared_ptr<Scene> (
                                    new LightScene(sceneNr, active, value, config));
                            break;
                        }
                        case MOTOR_SCENE:
                        {
                            pScene = boost::shared_ptr<Scene> (
                                    new MotorScene(sceneNr, active, value, config));
                            break;
                        }
                        case JALUSIE_SCENE:
                        {
                            pScene = boost::shared_ptr<Scene> (
                                    new JalusieScene(sceneNr, active, value, angle, config));
                            break;
                        }
                    }
                    res->addScene(pScene);
                }
            }
        }

        json_object* parametertable;
        if (json_object_object_get_ex(json_request, "parameter", &parametertable))
        {
            for (int i = 0; i < json_object_array_length(parametertable); ++i)
            {
                json_object* parameter = json_object_array_get_idx(parametertable, i);
                if (parameter != NULL)
                {
                    int bank = 0;
                    if (json_object_object_get_ex(parameter, "bank", &obj))
                    {
                        bank = json_object_get_int(obj);
                    }

                    int offset = 0;
                    if (json_object_object_get_ex(parameter, "offset", &obj))
                    {
                        offset = json_object_get_int(obj);
                    }

                    int value = 0;
                    if (json_object_object_get_ex(parameter, "value", &obj))
                    {
                        value = json_object_get_int(obj);
                    }

                    int mask = 0;
                    if (json_object_object_get_ex(parameter, "mask", &obj))
                    {
                        mask = json_object_get_int(obj);
                    }

                    boost::shared_ptr<ParameterSpec> pParam(new ParameterSpec());
                    pParam->bank = bank;
                    pParam->offset = offset;
                    pParam->value = value;
                    pParam->mask = mask;

                    res->addParameter(pParam);
                }
            }
        }
    }
    json_object_put(json_request);
    json_tokener_free(tok);

    return res;
}

}
