/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __STORAGE_H__
#define __STORAGE_H__

#include <boost/shared_ptr.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <list>

#include "hwdesc.h"

/// \file storage.h
namespace dSA
{

/// This class stores information about all discovered digitalSTROM devices.
///
/// We have a number of dS-Meters, each dS-Meter has a number of associated
/// devices or "Klemmen". The "Klemmen" are stored within each dSM and are
/// managed by the dSM class, however the Storage class also offers convenience
/// functions for working with devices.
/// We also have a number of "other" devices on the dS485
/// bus; for example a dSS will also show up in the bus members list. We can't
/// really do anything with those, but we will simply track their presence and
/// show them in our list.
///
/// This class is thread safe.
class Storage
{
public:
    static Storage *getInstance();

    void addDSM(boost::shared_ptr<dSM> dsm);
    void addDevice(boost::shared_ptr<Device> device);

    void removeDSM(dsuid_t dsuid);
    void removeDevice(dsuid_t);
    void removeDevice(dsuid_t, dsuid_t dsm_dsuid);
    void removeDevice(uint16_t device_id, dsuid_t dsm_dsuid);

    boost::shared_ptr<const HWBase> get(dsuid_t dsuid) const;
    boost::shared_ptr<dSM> getDSM(dsuid_t dsuid) const;
    boost::shared_ptr<Device> getDevice(dsuid_t dsuid) const;
    boost::shared_ptr<Device> getDevice(dsuid_t dsuid,
                                              dsuid_t dsm_dsuid) const;
    boost::shared_ptr<Device> getDevice(uint16_t deviceid,
                                              dsuid_t dsm_dsuid) const;

    boost::shared_ptr<std::list<dsuid_t> > getDSMList() const;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
        getDevices(bool onlyActive = true) const;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
        getDevices(dsuid_t dsm_dsuid, bool onlyActive = true) const;

    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                 getDeviceList(dsuid_t dsuid, bool onlyActive = true) const;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                 getDeviceList(dsuid_t dsm_dsuid, uint16_t device_id) const;
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                 getDeviceList(uint16_t zone, bool onlyActive = true) const;
protected:
    Storage();

private:
    static Storage* m_instance;
    mutable boost::recursive_mutex m_mutex;
};

};

#endif//__STORAGE_H__
